//The Loot Marker Handler watches for spawning items, checks the rules, then possibly adds a marker to the map
class LootMarkerHandler : EventHandler
{
	MarkerRuleHandler MarkerRules;
	
	protected array<LootMarker> AttachedMarker;
	protected array<actor> AttachedActor;
	protected array<string> AttachedActorName; //track original actor name for debugging
	
	override void OnRegister()
	{
		//console.printf("Initialize LootMarkers");//DEBUG
		
		MarkerRules = MarkerRuleLoader.GetRules();
	}
	
	override void WorldThingSpawned(WorldEvent e)
	{
		if (!e.thing) return;

		//DEBUG
		// string thisTag = e.thing.getTag();
		// string thisClassName = e.thing.getClassName();
		// string thisDisplayName = thisTag ~== thisClassName ? thisTag : thisClassName.." ["..thisTag.."]";
		// console.printf((MarkerRules.CheckRules(e.thing) ? "+" : "-").."Spawned: "..thisDisplayName);

		if (CheckMarkRules(e.thing)) AttachMarkerTo(e.thing);
	}
	
	override void ConsoleProcess(ConsoleEvent e){
		if (e.Name ~== "LootMarkersList")
		{
			PrintList();
		}
	}

	override void NetworkProcess(ConsoleEvent e){
		if (e.Name ~== "LootMarkersRulesRefreshed")
		{
			RefreshMarkers();
		}
	}

	static LootMarkerHandler GetInstance()
	{
		let h = LootMarkerHandler(EventHandler.Find('LootMarkerHandler'));

		if (h == null)
		{
			ThrowAbortException("Could not find the instance of LootMarkerHandler.");
		}

		return h;
	}
	
	//Check the rules to see if this actor should be marked
	bool CheckMarkRules(actor a)
	{
		if (!a) return false;
		if (a is "MapMarker") return false; //hardcode this to prevent loops
		return (MarkerRules.CheckRules(a));
	}
	
	//Spawns and attaches a loot marker to the actor without running any checks
	LootMarker AttachMarkerTo(actor attachee)
	{
		if (!attachee) return null;
		
		//silently ignore these
		if (attachee is "LootMarker") return null;

		LootMarker marker = LootMarker(actor.spawn("LootMarker",attachee.pos));;

		marker.AttachedTo = attachee;
		marker.Args[2] = 1; //Autoscale on map
		
		AttachedActor.push(attachee);
		AttachedActorName.push(attachee.GetCharacterName());
		AttachedMarker.push(marker);

		//Check filter
		marker.MarkerFiltered = MarkerRules.ShouldBeFiltered(attachee);

		return marker;
	}
	
	//destroy a marker and remove it's relationship from the list
	//return is success
	bool DestroyMarkerByIndex(int IndexNumber)
	{
		bool ret = false;
		
		if (0 <= IndexNumber && IndexNumber < AttachedMarker.size())
		{
			AttachedMarker[IndexNumber].Destroy();
			AttachedMarker.Delete(IndexNumber);
			AttachedActor.Delete(IndexNumber);
			AttachedActorName.Delete(IndexNumber);
			
			ret = true;
		}
		
		return ret;
	}
	
	//destroy a marker and remove it's relationship from the list
	//return is success
	bool DestroyMarker(LootMarker marker)
	{
		if (!marker) return false;
	
		uint i = 0;
		bool FoundOne = false;
		
		//Shouldn't need to do this loop more than once, but just in case
		do
		{
			i = AttachedMarker.find(marker);
			//console.printf("LootMarkers: Remove %d / %d", i, AttachedMarker.size());//DEBUG
			FoundOne |= DestroyMarkerByIndex(i);
		}
		while (i < AttachedMarker.size());
		
		return FoundOne;
	}
	
	//destroy a marker and remove it's relationship from the list
	//return is success
	bool DestroyMarkerFrom(actor attachee)
	{
		if (!attachee) return false;

		uint i = 0;
		bool FoundOne = false;
		
		//Shouldn't need to do this loop more than once, but just in case
		Do
		{
			i = AttachedActor.find(attachee);
			FoundOne |= DestroyMarkerByIndex(i);
		}
		while (i < AttachedActor.size());
		
		return FoundOne;
	}

	//Find a marker attahed to the given actor
	LootMarker FindAttachedMarker(actor attachee)
	{
		if (!attachee) return null;

		LootMarker marker = null;

		uint i = AttachedActor.find(attachee);
		if (i < AttachedMarker.size())
		{
			marker = AttachedMarker[i];
			
			//this shouldn't happen but just in case, house keeping
			if (!marker) DestroyMarkerByIndex(i);
		}
		
		return marker;
	}
	
	//checks all actors, adds missing markers, removes inactive markers, active markers are unchanged
	//only need to call this if the rules have changed
	void RefreshMarkers()
	{
		ThinkerIterator it = ThinkerIterator.Create("Actor");
		Actor mo;
		while (mo = Actor(it.Next()))
		{
			LootMarker lm;
			if (MarkerRules.CheckRules(mo))
			{	//This actor should automatically have a marker
				//Look for a marker, spawn one if not found
				lm = FindAttachedMarker(mo);
				if (!lm) lm = AttachMarkerTo(mo);
			}
			else
			{	//This actor should not automatically have a marker
				lm = FindAttachedMarker(mo);
				//leave active markers alone
				if (lm && !lm.MarkerActive) DestroyMarker(lm);
			}

			//update filter
			if (lm) lm.MarkerFiltered = MarkerRules.ShouldBeFiltered(mo);
		}
	}

	//for debugging
	clearscope void PrintList()
	{
		uint iMax = max(AttachedActor.size(),AttachedMarker.size());
		console.printf("LootMarker List %d / %d", AttachedActor.size(),AttachedMarker.size());

		string aName, mName, bActive, bFiltered;
		
		for (uint i = 0; i < iMax; i++)
		{
			aName="[NULL]";
			mName="[NULL]";
			bActive="-";
			bFiltered="-";
			
			if (i < AttachedActor.size())
			{
				actor a = AttachedActor[i];
				
				if (a) aName=a.GetCharacterName();
			}
			
			if (aName != AttachedActorName[i]) aName = aName.."["..AttachedActorName[i].."]";

			if (i < AttachedMarker.size())
			{
				LootMarker m = AttachedMarker[i];
				
				if (m) 
				{
					mName=m.GetCharacterName();
					if (m.MarkerActive) bActive="A";
					if (m.MarkerFiltered) bFiltered="F";
				}
				else
				{
					bActive="?";
					bFiltered="?";
				}
			}
			
			console.printf("%s%s %s=%s",bActive,bFiltered,aName,mName);
		}
	}
}

//The Loot Marker is the actual map marker object. It watches it's attachee for changes and whether the player can see it
class LootMarker : MapMarker
{
	//An active marker fits the rules and has been seen by the player
	bool MarkerActive;
	//A filtered marker may or may not have been seen by the player, but has been filtered by the player
	bool MarkerFiltered;
	
	default
	{
		+DORMANT;
		// +NOINTERACTION;
		// +NOBLOCKMAP;
	}

	states
	{
		spawn:
			UNKN A -1;
			stop;
	}

	actor AttachedTo;
	
	//MapMarker class has the activate/deactivate swapped, so I swap them back
	override void Activate(actor activator){super.deactivate(activator);}
	override void Deactivate(actor activator){super.activate(activator);}

	override void Tick(void)
	{
		Super.tick();

		
		bool ShouldBeActive, ShouldBeMoved;
		[ShouldBeActive, ShouldBeMoved] = ShouldBeActiveAndMoved();

		//Determine if marker needs to be activated or deactivated
		if (!MarkerActive)
		{
			if (!AttachedTo)
			{
				//got no reason to live
				LootMarkerHandler.GetInstance().DestroyMarker(self);
			}
			else if (ShouldBeActive)
			{
				MarkerActive = true;
			}
		}
		else
		{
			if (!ShouldBeActive)
			{
				MarkerActive = false;
			}
		}
		
		//Update marker visuals if it is seen
		if (ShouldBeActive && AttachedTo && CanPlayersSee(AttachedTo))
		{
			sprite = AttachedTo.sprite;
			frame = AttachedTo.frame;
			translation = AttachedTo.translation;
			if (radius != AttachedTo.radius || height != AttachedTo.height)
				A_SetSize(AttachedTo.radius,AttachedTo.height);
			Scale.x = AttachedTo.Scale.x;
			Scale.y = AttachedTo.Scale.y;
		}
		if (ShouldBeMoved)
		{
			SetOrigin(AttachedTo.pos, true);
		}
		
		//Actually activate or not depending on filtration
		if (bDormant)
		{
			if (MarkerActive && !MarkerFiltered) Activate(null);
		}
		else
		{
			if (!MarkerActive || MarkerFiltered) Deactivate(null);
		}
		
	}
	
	//internal use, tells if the marker should be active and/or moved
	virtual bool, bool ShouldBeActiveAndMoved()
	{
		bool ShouldBeActive, ShouldBeMoved;
		
		if (!MarkerActive)
		{
			if (!AttachedTo)
			{
				// Nothing to see here
				ShouldBeActive=false;
				ShouldBeMoved=false;
			}
			else
			{
				// Maybe something to see here
				//ShouldBeActive=!AttachedTo.CheckIfSeen();
				ShouldBeActive=CanPlayersSee(AttachedTo);
				ShouldBeMoved=ShouldBeActive;
			}
		}
		else
		{
			//attached item has been moved or removed
			bool Moved = (!AttachedTo || pos != AttachedTo.pos);
			
			if (!Moved)
			{
				// No change
				ShouldBeActive=true;
				ShouldBeMoved=false;
			}
			else
			{
				//if (!AttachedTo || AttachedTo.CheckIfSeen())
				if (!AttachedTo || !CanPlayersSee(AttachedTo))
				{
					// See a missing item
					//ShouldBeActive=CheckIfSeen();
					ShouldBeActive=!CanPlayersSee(self);
					ShouldBeMoved=false;
				}
				else
				{
					// See a moved item
					ShouldBeActive=true;
					ShouldBeMoved=true;
				}
			}
			
			
		}
		
		return ShouldBeActive, ShouldBeMoved;
	}

	//Use this to attach a marker to a specific actor
	static LootMarker AttachOneTo(actor attachee)
	{
		return LootMarkerHandler.GetInstance().AttachMarkerTo(attachee);
	}
	
	//Use this to find a marker attached to a specific actor
	static LootMarker FindAttached(actor attachee)
	{
		return LootMarkerHandler.GetInstance().FindAttachedMarker(attachee);
	}
	
	//Use this to remove a marker attached to a specific actor
	static bool RemoveAttachedFrom(actor attachee)
	{
		return LootMarkerHandler.GetInstance().DestroyMarkerFrom(attachee);
	}

	//a crude function, not perfect, but good enough for now
	static bool CanPlayerSee(int PlayerNum, actor CheckTarget)
	{
		if (CheckHitboxIntersection(players[PlayerNum].mo,CheckTarget)) return true;
		
		if (!players[PlayerNum].mo.CheckSight(CheckTarget, SF_IGNOREVISIBILITY | SF_SEEPASTSHOOTABLELINES)) return false;

		double PlayAng = players[PlayerNum].mo.angle;  
		double AngleToMe = players[PlayerNum].mo.AngleTo(CheckTarget);
		double AngleDiff = AbsAngle(AngleToMe, PlayAng);
		double PlayFovRange = players[PlayerNum].fov / 2;

		return AngleDiff < PlayFovRange;
	}

	//Calls CanPlayerSee for all players
	static bool CanPlayersSee(actor CheckTarget)
	{
		for (int i = 0; i < players.size(); i++)
		{
			if (players[i].mo && CanPlayerSee(i, CheckTarget)) return true;
		}
		return false;
	}
	
//  **************************************************************************
//  Code provided by user inkoalawetrust from ZDoom forums
//  **************************************************************************

	//Checks if two hitboxes are clipping into one another.
	Static ClearScope Bool CheckHitboxIntersection (Actor A, Actor B)
	{
		If (!A || !B)
			Return False;
		
		Vector3 AMin, AMax;
		Vector3 BMin, BMax;
		[AMin, AMax] = GetHitBoxBoundaries(A);
		[BMin, BMax] = GetHitBoxBoundaries(B);
		
		Return (
		AMin.X <= BMax.X &&
		AMax.X >= BMin.X &&
		AMin.Y <= BMax.Y &&
		AMax.Y >= BMin.Y &&
		AMin.Z <= BMax.Z &&
		AMax.Z >= BMin.Z
		);
	}
	
	//Returns the minimum and maximum (Respectively) XYZ boundaries of a hitbox in 2 Vector3s.
	Static ClearScope Vector3, Vector3 GetHitboxBoundaries (Actor Mo)
	{
		Vector3 MobjMin = (Mo.Pos.X-Mo.Radius,Mo.Pos.Y-Mo.Radius,Mo.Pos.Z);
		Vector3 MobjMax = (Mo.Pos.X+Mo.Radius,Mo.Pos.Y+Mo.Radius,Mo.Pos.Z+Mo.Height);
		
		Return MobjMin, MobjMax;
	}

//  **************************************************************************
//  End of code provided by user inkoalawetrust from ZDoom forums
//  **************************************************************************
}
