//Use a Static Event Handler to load the rules because they don't need to be reloaded on every map
class MarkerRuleLoader : StaticEventHandler
{
	MarkerRuleHandler MarkerRules;

	private bool _LootMarkersAutoInv;
	private bool _LootMarkersAutoOther;
	private string _LootMarkersOtherClasses;
	private bool _LootMarkersFilterOn;
	private string _LootMarkersFilterClasses;
	
	override void OnRegister()
	{
		console.printf("Initialize LootMarkers rules");
		
		MarkerRules=new("MarkerRuleHandler");
		MarkerRules.RebuildPrecheckList(); //read rules from user settings
		MarkerRules.ReadRulesFromLump("LOOTMARK"); //rest of rules are read from lump(s)
		MarkerRules.RebuildFilterList();//read filters from user settings
		//MarkerRules.PrintRules();//DEBUG

		_LootMarkersAutoInv = LootMarkersAutoInv;
		_LootMarkersAutoOther = LootMarkersAutoOther;
		_LootMarkersOtherClasses = LootMarkersOtherClasses;
		_LootMarkersFilterOn = LootMarkersFilterOn;
		_LootMarkersFilterClasses = LootMarkersFilterClasses;
	}
	
	override void WorldTick()
	{

		//check if need to refresh user rules
		if (
			_LootMarkersAutoInv != LootMarkersAutoInv ||
			_LootMarkersAutoOther != LootMarkersAutoOther ||
			_LootMarkersOtherClasses != LootMarkersOtherClasses ||
			_LootMarkersFilterOn != LootMarkersFilterOn ||
			_LootMarkersFilterClasses != LootMarkersFilterClasses
		)
		{
			MarkerRules.RebuildPrecheckList();
			MarkerRules.RebuildFilterList();
			EventHandler.SendNetworkEvent("LootMarkersRulesRefreshed");
			_LootMarkersAutoInv = LootMarkersAutoInv;
			_LootMarkersAutoOther = LootMarkersAutoOther;
			_LootMarkersOtherClasses = LootMarkersOtherClasses;
			_LootMarkersFilterOn = LootMarkersFilterOn;
			_LootMarkersFilterClasses = LootMarkersFilterClasses;
		}

	}

	override void ConsoleProcess(ConsoleEvent e)
	{
		if (e.Name ~== "LootMarkersRules")
		{
			MarkerRules.PrintRules();
		}
	}
	
	static MarkerRuleHandler GetRules()
	{
		let h = MarkerRuleLoader(StaticEventHandler.Find('MarkerRuleLoader'));

		if (h == null)
		{
			ThrowAbortException("Could not find the instance of MarkerRuleLoader.");
		}

		return h.MarkerRules;
	}
}

//The Marker Rule Hander maintains the list of rules and filters
class MarkerRuleHandler
{
	private array<bool> RuleIsInclusive;
	private array<string> RuleText;

	private array<string> PrecheckClassList;
	
	private array<string> FilterClassList;
	
	//Check that arrays have not gone out of sync
	void SanityCheck()
	{
		//this should never happen
		if (RuleIsInclusive.size() != RuleText.size())
		{
			PrintRules();
			ThrowAbortException("Marker Rule List is corrupted");
			
			//or just start over
			RuleIsInclusive.clear();
			RuleText.clear();
		}
	}
	
	uint AddRuleToList(bool InclusiveRule, string MatchText)
	{
		SanityCheck();
		
		RuleIsInclusive.push(InclusiveRule);
		uint i = RuleText.push(MatchText);
		
		return i;
	}
	
	//try to add a line as a rule. Line should be like "(include|exclude)=(match text)"
	bool TryAddLineAsRule(string RuleLine)
	{
		if (!RuleLine || RuleLine ~== "" || RuleLine.left(2) ~== "//") return false;
		
		int splitpoint = RuleLine.IndexOf("=");
		if (splitpoint < 0) return false;
		if (splitpoint >= RuleLine.length()-1) return false;
		//console.printf("rule [split,len]=[%d,%d] %s",splitpoint,RuleLine.length(), RuleLine);//DEBUG

		string RuleKey=RuleLine.left(splitpoint);
		string RuleVal=RuleLine.mid(splitpoint+1,int.max);
		
		if (RuleKey~=="include")
		{
			AddRuleToList(true, RuleVal);
			return true;
		}
		if (RuleKey~=="exclude")
		{
			AddRuleToList(false, RuleVal);
			return true;
		}
		return false;
	}
	
	//reads rules from specified lump and successive lumps
	void ReadRulesFromLump(string LumpName)
	{
		int lump = -1, next = 0;
		while ((lump = Wads.FindLump(LumpName, next, Wads.GLOBALNAMESPACE)) >= 0)
		{
			next = lump + 1;
			string thisLumpName=Wads.GetLumpFullName(lump);
			//Console.Printf("[MarkerRules]Reading lump: %s", thisLumpName);//DEBUG

			let content = Wads.ReadLump(lump);
			Array<string> lines;
			content.Split(lines, "\n", TOK_SKIPEMPTY);

			for (uint i = 0; i < lines.Size(); i++)
			{
				lines[i].StripRight("\r"); //trim CR leftovers from CRLF files
			
				if (lines[i] ~== "" || lines[i].Left(2) ~== "//") continue;

				if (!TryAddLineAsRule(lines[i]))
				{
					console.printf("[ERROR]Marker Rules: Bad line [%d] in %s: %s", i + 1, thisLumpName, lines[i]);
				}
			}
		}
	}

	void RebuildPrecheckList()
	{
		PrecheckClassList.clear();
		if (LootMarkersAutoOther) LootMarkersOtherClasses.split(PrecheckClassList, ",", TOK_SKIPEMPTY);
		if (LootMarkersAutoInv) PrecheckClassList.Insert(0,"Inventory");
	}
	
	void RebuildFilterList()
	{
		FilterClassList.clear();
		if (LootMarkersFilterOn) LootMarkersFilterClasses.split(FilterClassList, ",", TOK_SKIPEMPTY);
		//console.printf("Marker filters: %d", FilterClassList.size());//DEBUG
	}
	
	//checks the user supplied rules only
	bool PrecheckRules(actor a)
	{
		for (int i = 0; i < PrecheckClassList.size(); i++)
		{
			if (a is PrecheckClassList[i]) return true;
		}
		return false;
	}
	
	//check if given actor matches rules for include or exclude
	bool CheckRules(actor a)
	{
		if (!a) return false;
		
		SanityCheck();
		
		//bool ShouldInclude = (LootMarkersAutoInv && a is "Inventory");
		bool ShouldInclude = PrecheckRules(a);
		
		string thisRule;
		string thisTag=a.GetTag();
		string thisClass=a.GetClassName();

		for (int i = 0; i < RuleText.size(); i++)
		{
			//ignore rules that don't matter
			if (ShouldInclude && RuleIsInclusive[i]) continue; 
			if (!ShouldInclude && !RuleIsInclusive[i]) continue; 
			
			thisRule=RuleText[i];
			bool MatchesRule = thisTag ~== thisRule || thisClass ~== thisRule || (a is thisRule);
			
			if (MatchesRule) ShouldInclude = RuleIsInclusive[i];
		}
		
		return ShouldInclude;
	}
	
	//check if given actor matches any filter
	bool ShouldBeFiltered(actor a)
	{
		string thisFilterItem;
		string thisTag=a.GetTag();
		string thisClass=a.GetClassName();
		bool matchesFilter;

		for (int i = 0; i < FilterClassList.size(); i++)
		{
			thisFilterItem=FilterClassList[i];
			matchesFilter = thisTag ~== thisFilterItem || thisClass ~== thisFilterItem || (a is thisFilterItem);
			if (matchesFilter) return true;
		}
		return false;
	}
	
	uint PrintRules()
	{
		console.printf("Marker Rules "..RuleIsInclusive.size().."/"..RuleText.size());
		
		for (int i = 0; i < PrecheckClassList.size(); i++)
		{
			console.printf("[%d]UserInclude=%s", i, PrecheckClassList[i]);
		}
		
		//Arrays *should* be the same size, but just in case they're not
		int iMax = max(RuleIsInclusive.size(),RuleText.size());
		
		for (int i = 0; i < iMax; i++)
		{
			string RuleKey=(i < RuleIsInclusive.size()) ? (RuleIsInclusive[i] ? "include" : "exclude") : "[Undefined]";
			string RuleVal=(i < RuleText.size()) ? RuleText[i] : "[Undefined]";
			
			console.printf("["..i.."]"..RuleKey.."="..RuleVal);
		}
		
		return iMax;
	}
}

